#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "ipcevent.h"

/* In sending a message, the sequence is: Process sends the message,
   waits for an IPC_ACK response, and then, optionally, waits for the
   reply.  While waiting for a reply, callbacks may occur ant will be
   processed. The only thing complicated here is the collision
   protocol: it both processes send at the same time, then each
   receives an IPC_SEND message when it expects IPC_ACK.  Collisions
   like this are handled by giving one process priority. The process
   with priority discards the IPC_SEND message it receives, waits for
   a reply, if requested, and then waits for the interrupted send
   message to be resent.  The other process knows its IPC_SEND message
   will be discarded; it sends IPC_ACK, processes the message IPC_SEND
   message it received and sends IPC_REPLY, if requested, and the
   resends its original message. I think this is sufficient.
   Currently priority is assigned at startup and stays fixed.  It
   would be possible to switch priority on each collision.  It would
   also be possible for the process with priority to store the
   IPC_SEND message it ignores to avoid having it be retransmitted.
   Probably neither is worth bothering with. */

struct ipc_event_sink_t {
  pid_t childpid;
  int readfd, writefd;
  ipc_bool_t priority;
  ipc_event_handler_t handler;
};

struct ipc_event_t {
  ipc_size_t size;
  void *data;
};

static int SinkRead(ipc_event_sink_t sink, void *buf, size_t size)
{
  if (read(sink->readfd, buf, size) != size)
    return -1;
  return 0;
}

static int SinkWrite(ipc_event_sink_t sink, void *buf, size_t size)
{
  if (write(sink->writefd, buf, size) != size)
    return -1;
  return 0;
}

int IPC_StartServer(char *progpath, char *argv[],
		    ipc_event_sink_t *psink, ipc_event_handler_t handler)
{
  int p2c[2], c2p[2];
  pid_t childpid;

  static int inited = 0;
  if (! inited)
    inited = 1;
  else
    return -1;

  if (pipe(p2c) != 0)
    return -1;
  if (pipe(c2p) != 0)
    return -1;

  switch ((childpid = fork())) {
  case -1:  /* fork failed */
    return -1;
  case 0:   /* child */
    /***** do some error checking here */
    dup2(p2c[0], STDIN_FILENO);
    dup2(c2p[1], STDOUT_FILENO);
    close(p2c[0]);
    close(p2c[1]);
    close(c2p[0]);
    close(c2p[1]);
    execv(progpath, argv);
    perror("exec failed");
    exit(1);
  default:  /* parent */
    close(p2c[0]);
    close(c2p[1]);
    
    /**** blow away the child and reset inited on errors? */
    if ((*psink = malloc(sizeof(struct ipc_event_sink_t))) == NULL)
      return -1;

    (*psink)->priority = 1;
    (*psink)->handler = handler;
    (*psink)->readfd = c2p[0];
    (*psink)->writefd = p2c[1];
    (*psink)->childpid = childpid;
  }
  return 0;
}

int IPC_InitServer(ipc_event_sink_t *psink, ipc_event_handler_t handler)
{
  static int inited = 0;

  if (! inited)
    inited = 1;
  else
    return -1;

  if ((*psink = malloc(sizeof(struct ipc_event_sink_t))) == NULL)
    return -1;
  
  (*psink)->priority = 0;
  (*psink)->handler = handler;
  (*psink)->readfd = STDIN_FILENO;
  (*psink)->writefd = STDOUT_FILENO;
  (*psink)->childpid = 0;

  return 0;
}

int IPC_Disconnect(ipc_event_sink_t sink)
{
  close(sink->readfd);
  close(sink->writefd);
  /**** kill and or wait on server death?? */
  return 0;
}

int IPC_SinkFD(ipc_event_sink_t sink, int *pfd)
{
  *pfd = sink->readfd;
  return 0;
}

int IPC_CreateEvent(ipc_event_t *pevent)
{
  if ((*pevent = malloc(sizeof(struct ipc_event_t))) == NULL)
    return -1;
  (*pevent)->size = 0;
  (*pevent)->data = NULL;
  return 0;
}

int IPC_DisposeEvent(ipc_event_t event)
{
  free(event->data);
  free(event);
  return 0;
}

int IPC_PutEventData(ipc_event_t event, void *data, ipc_size_t size)
{
  if (event->data != NULL) {
    free(event->data);
    event->data = NULL;
    event->size = 0;
  }
  if (size > 0) {
    if ((event->data = malloc(size)) == NULL)
      return -1;
    memcpy(event->data, data, size);
    event->size = size;
  }
  return 0;
}

int IPC_AppendEventData(ipc_event_t event, void *data, ipc_size_t size)
{
  if (size > 0) {
    void *newdata;
    if ((newdata = realloc(event->data, event->size + size)) == NULL)
      return -1;
    event->data = newdata;
    memcpy(((char *) event->data) + event->size, data, size);
    event->size += size;
  }
  return 0;
}

int IPC_GetEventData(ipc_event_t event, void **pdata, ipc_size_t *psize)
{
  *pdata = event->data;
  *psize = event->size;
  return 0;
}

static int HandleSendEvent(ipc_event_sink_t sink)
{
  ipc_size_t size;
  ipc_event_type_t type;
  ipc_event_t event;
  ipc_bool_t wait;

#ifdef DEBUG
  fprintf(stderr, "IPC_SEND received; sending IPC_ACK\n");
#endif
  type = IPC_ACK;
  if (SinkWrite(sink, &type, sizeof(type)) != 0)
    return -1;

  if (SinkRead(sink, &wait, sizeof(wait)) != 0)
    return -1;
  if (SinkRead(sink, &size, sizeof(size)) != 0)
    return -1;
  if (IPC_CreateEvent(&event) != 0)
    return -1;
  if (size > 0) {
    if ((event->data = malloc(size)) == NULL) {
      IPC_DisposeEvent(event);
      return -1;
    }
    if (SinkRead(sink, event->data, size) != 0) {
      IPC_DisposeEvent(event);
      return -1;
    }
  }
  else
    event->data = NULL;
  event->size = size;

  if (sink->handler != NULL)
    sink->handler(event, wait);

  if (wait) {
    type = IPC_REPLY;
    size = event->size;
    if (SinkWrite(sink, &type, sizeof(type)) != 0)
      return -1;
    if (SinkWrite(sink, &size, sizeof(size)) != 0)
      return -1;
    if (SinkWrite(sink, event->data, event->size) != 0)
      return -1;
  }

  IPC_DisposeEvent(event);
  return 0;
}

static int SendEventData(ipc_event_sink_t sink, ipc_event_t event,
			 ipc_bool_t wait)
{
  ipc_size_t size = event->size;

  if (SinkWrite(sink, &wait, sizeof(wait)) != 0)
    return -1;
  if (SinkWrite(sink, &size, sizeof(size)) != 0)
    return -1;
  if (SinkWrite(sink, event->data, size) != 0)
    return -1;
  return 0;
}

static int WaitEventReply(ipc_event_sink_t sink, ipc_event_t event)
{
  ipc_size_t size;
  ipc_event_type_t type;
  void *data;

  if (SinkRead(sink, &type, sizeof(type)) != 0)
    return -1;

  while (type == IPC_SEND) {
    if (HandleSendEvent(sink) != 0)
      return -1;
    if (SinkRead(sink, &type, sizeof(type)) != 0)
      return -1;
  }
  if (type != IPC_REPLY)
    return -1;
      
  if (SinkRead(sink, &size, sizeof(size)) != 0)
    return -1;
  if (size > 0) {
    if ((data = malloc(size)) == NULL)
      return -1;
    if (SinkRead(sink, data, size) != 0) {
      free(data);
      return -1;
    }
  }
  else
    data = NULL;
  if (event->data != NULL)
    free(event->data);
  event->size = size;
  event->data = data;

  return 0;
}


static int SendEvent(ipc_event_sink_t sink, ipc_event_t event, ipc_bool_t wait)
{
  if (SendEventData(sink, event, wait) != 0)
    return -1;
  if (wait)
    if (WaitEventReply(sink, event) != 0)
      return -1;
  return 0;
}

static int FlushSendEvent(ipc_event_sink_t sink)
{
  ipc_bool_t wait;
  ipc_size_t size;
  void *data;

  if (SinkRead(sink, &wait, sizeof(wait)) != 0)
    return -1;
  if (SinkRead(sink, &size, sizeof(size)) != 0)
    return -1;
  if (size > 0) {
    if ((data = malloc(size)) == NULL)
      return -1;
    if (SinkRead(sink, data, size) != 0) {
      free(data);
      return -1;
    }
    free(data);
  }
  return 0;
}

int IPC_SendEvent(ipc_event_sink_t sink, ipc_event_t event, ipc_bool_t wait)
{
  ipc_event_type_t type = IPC_SEND;
  ipc_bool_t collision;

  if (SinkWrite(sink, &type, sizeof(type)) != 0)
    return -1;

  if (SendEventData(sink, event, wait) != 0)
    return -1;

  if (SinkRead(sink, &type, sizeof(type)) != 0)
    return -1;
  if (type == IPC_SEND)
    collision = 1;
  else
    collision = 0;

  if (collision) {
    fprintf(stderr, "Collision!\n");
    if (sink->priority) {
      FlushSendEvent(sink);
      if (SinkRead(sink, &type, sizeof(type)) != 0)
	return -1;
      if (type != IPC_ACK)
	return -1;
      if (wait)
	if (WaitEventReply(sink, event) != 0)
	  return -1;
      return HandleSendEvent(sink);
    }
    else {
      if (HandleSendEvent(sink) != 0)
	return -1;
      if (SinkRead(sink, &type, sizeof(type)) != 0)
	return -1;
      if (type != IPC_ACK)
	return -1;
      return SendEvent(sink, event, wait);
    }
  }
  else if (wait)
    return WaitEventReply(sink, event);
  else
    return 0;
}

int IPC_HandleEvent(ipc_event_sink_t sink)
{
  ipc_event_type_t type;

  if (SinkRead(sink, &type, sizeof(type)) != 0 || type != IPC_SEND)
    return -1;

  return HandleSendEvent(sink);
}
