/**** add signal handling--make sure connection is closed */

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <tcl.h>
#include <tk.h>

#include "ipcevent.h"
#include "ipctk.h"

static char *ipctk_eval_cmd = "ipceval";
static char *ipctk_call_cmd = "ipccall";
static char *ipctk_call_lang_cmd = "ipccalllang";

static ipc_event_sink_t sink = NULL;
static Tcl_Interp *Tcl_interp;

static void ipcproc(ClientData data, int mask)
{
  if (TCL_READABLE & mask) {
    if (IPC_HandleEvent(sink) != 0) {
      fprintf(stderr, "event handler failed, exiting\n");
      Tcl_Exit(1);
    }
  }
}

static void ipcexit(ClientData data)
{
  if (sink != NULL) {
    fprintf(stderr, "disconnecting ...");
    IPC_Disconnect(sink);
    fprintf(stderr, "done\n");
    sink = NULL;
  }
}
  
#define BSIZE 512

static void handler(ipc_event_t event, ipc_bool_t want_reply)
{
  char *cmd;
  ipc_size_t size;
  ipctk_result_code_t code;

  IPC_GetEventData(event, (void **) &cmd, &size);
  
  if (Tcl_Eval(Tcl_interp, cmd) == TCL_ERROR)
    code = IPCTK_ERROR;
  else
    code = IPCTK_OK;

  if (want_reply) {
    char *result = Tcl_interp->result;
    IPC_PutEventData(event, &code, sizeof(code));
    IPC_AppendEventData(event, result, strlen(result) + 1);
  }
}

/***** same except code-- code, count, arg strings */
/***** allow create data if total size is known */
static int do_ipc_call(ClientData clientData,
		       Tcl_Interp *interp,
		       int argc,
		       char *argv[],
		       ipctk_event_type_t which)
{
  ipc_event_t ce;
  int i, result;
  ipctk_call_header_t hdr;
  char *data;
  ipc_size_t size;
  ipctk_result_code_t code;

  hdr.type = which;
  hdr.argc = argc;

  if (IPC_CreateEvent(&ce) != 0) {
    Tcl_SetResult(interp, "IPC Create Event failed", TCL_STATIC);
    return TCL_ERROR;
  }
  if (IPC_PutEventData(ce, &hdr, sizeof(hdr)) != 0) {
    Tcl_SetResult(interp, "IPC Put Event Data failed", TCL_STATIC);
    IPC_DisposeEvent(ce);
    return TCL_ERROR;
  }
  for (i = 0; i < argc; i++) {
    if (IPC_AppendEventData(ce, argv[i], strlen(argv[i]) + 1) != 0) {
      Tcl_SetResult(interp, "IPC Append Event Data failed", TCL_STATIC);
      IPC_DisposeEvent(ce);
      return TCL_ERROR;
    }
  }

  if (IPC_SendEvent(sink, ce, 1) != 0) {
    Tcl_SetResult(interp, "IPC Send Event failed", TCL_STATIC);
    IPC_DisposeEvent(ce);
    return TCL_ERROR;
  }

  if (IPC_GetEventData(ce, (void **) &data, &size) != 0) {
    Tcl_SetResult(interp, "IPC Get Event Data failed", TCL_STATIC);
    IPC_DisposeEvent(ce);
    return TCL_ERROR;
  }

  code = *((ipctk_result_code_t *) data);
  data += sizeof(ipctk_result_code_t);
  
  if (code == IPCTK_ERROR) {
    Tcl_SetResult(interp, data, TCL_VOLATILE);
    result = TCL_ERROR;
  }
  else
    result = TCL_OK;

  IPC_DisposeEvent(ce);
  return result;
}

static int ipceval(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc,
		   char *argv[])
{
  return do_ipc_call(clientData, interp, argc, argv, IPCTK_EVAL);
}

static int ipccall(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc,
		   char *argv[])
{
  return do_ipc_call(clientData, interp, argc, argv, IPCTK_CALL);
}

static int ipccalllang(ClientData clientData,
		       Tcl_Interp *interp,
		       int argc,
		       char *argv[])
{
  return do_ipc_call(clientData, interp, argc, argv, IPCTK_CALL_LANG);
}

static void error(char *sys, char *msg)
{
  fprintf(stderr, "Error [%s]: %s\n", sys, msg);
  Tcl_Exit(1);
}

static void tclerror(char *msg) { error("TCL", msg); }
static void ipcerror(char *msg) { error("IPC", msg); }

int main(int argc, char * argv[])
{
  int fd;

  if (argc > 1)
    ipctk_eval_cmd = argv[1];
  if (argc > 2)
    ipctk_call_cmd = argv[2];
  if (argc > 3)
    ipctk_call_lang_cmd = argv[3];

  /* Ignore interrupts so ^C in R process doesn't kill server, A
   possible alternative is to chenge the process group of the server
   with setpgid(0, 0).  There is a small race condition since the
   server is vulnerable until the signal handling is addressed.*/
  signal(SIGINT, SIG_IGN);

  Tcl_interp = Tcl_CreateInterp();

  if (Tcl_Init(Tcl_interp) == TCL_ERROR)
    tclerror(Tcl_interp->result);
  if (Tk_Init(Tcl_interp) == TCL_ERROR)
    tclerror(Tcl_interp->result);
  if (Tcl_Eval(Tcl_interp, "wm withdraw .") == TCL_ERROR) /* hide window */
    tclerror(Tcl_interp->result);

  Tcl_CreateCommand(Tcl_interp, ipctk_eval_cmd, ipceval, NULL, NULL);
  Tcl_CreateCommand(Tcl_interp, ipctk_call_cmd, ipccall, NULL, NULL);
  Tcl_CreateCommand(Tcl_interp, ipctk_call_lang_cmd, ipccalllang, NULL, NULL);


  if (IPC_InitServer(&sink, handler) != 0)
    ipcerror("connect failed");
  if (IPC_SinkFD(sink, &fd) != 0)
    ipcerror("getting sink fd failed");

  Tcl_CreateFileHandler(fd, TCL_READABLE, ipcproc, NULL);

  Tcl_CreateExitHandler(ipcexit, NULL);

  if (Tcl_Eval(Tcl_interp, "vwait ipc_quit") == TCL_ERROR)
    tclerror(Tcl_interp->result);

  Tcl_Exit(0);
  return 0;
}
