#include <sys/types.h>
#include <sys/time.h>
#include <Rinternals.h>
#include "R_ext/eventloop.h"
#include "R_ext/Parse.h"
#include "ipcevent.h"
#include "ipctk.h"

static ipc_event_sink_t sink;

static void handler(ipc_event_t event, ipc_bool_t want_reply)
{
    ipctk_call_header_t hdr;
    ipctk_result_code_t code;
    ipc_size_t size;
    char *data, *msg;

    IPC_GetEventData(event, (void **) &data, &size);
    hdr = *((ipctk_call_header_t *) data);
    data += sizeof(hdr);

    switch (hdr.type) {
    case IPCTK_EVAL:
	{
	    ParseStatus status;
	    int i;
	    SEXP text, expr, ans;

	    text = PROTECT(allocVector(STRSXP, hdr.argc - 1));
	    data += strlen(data) + 1;
	    for (i = 0 ; i < hdr.argc - 1; i++) {
		SET_STRING_ELT(text, i, mkChar(data));
		data += strlen(data) + 1;
	    }

	    expr = PROTECT(R_ParseVector(text, -1, &status));
	    if (status != PARSE_OK) {
		code = IPCTK_ERROR;
		msg = "parse error in R expression";
	    }
	    else {
		/* Note that expr becomes an EXPRSXP and hence we need
		   the loop below (a straight eval(expr, R_GlobalEnv)
		   won't work) */
		int n = length(expr);
		for(i = 0 ; i < n ; i++)
		    /******** need an exit trap here */
		    ans = eval(VECTOR_ELT(expr, i), R_GlobalEnv);
		code = IPCTK_OK;
		msg = "";
	    }
	    UNPROTECT(2);
	    break;
	}
    case IPCTK_CALL:
	{
	    int i;
	    SEXP expr, fun, last;

	    expr = R_NilValue;
	    data += strlen(data) + 1;
	    fun = (SEXP) strtoul(data, NULL, 16); /**** make a stable pointer*/
	    data += strlen(data) + 1;

	    PROTECT(expr = LCONS(fun, R_NilValue));
	    last = expr;
	    for (i = 0; i < hdr.argc - 2; i++) {
		SETCDR(last, LCONS(mkString(data), R_NilValue));
		data += strlen(data) + 1;
		last = CDR(last);
	    }
	    UNPROTECT(1);

	    PROTECT(expr = LCONS(install("try"), LCONS(expr, R_NilValue)));
	    /******** need an exit trap here */
	    eval(expr, R_GlobalEnv);
	    UNPROTECT(1);

	    code = IPCTK_OK;
	    msg = "";
	    break;
	}
    case IPCTK_CALL_LANG: 
	if (hdr.argc == 2) {
	    SEXP expr, env;
	    expr = (SEXP) strtoul(data, NULL, 16); data += strlen(data) + 1;
	    env  = (SEXP) strtoul(data, NULL, 16); data += strlen(data) + 1;
	    expr = LCONS(install("try"), LCONS(expr, R_NilValue));
	    eval(expr, env);
	    code = IPCTK_OK;
	    msg = "";
	}
	else {
	    code = IPCTK_ERROR;
	    msg = "wrong number of arguments for R_call_lang";
	}
	break;
    default:
	code = IPCTK_ERROR;
	msg = "unknown callback type";
    }
    IPC_PutEventData(event, &code, sizeof(code));
    if (code == IPCTK_ERROR)
	IPC_AppendEventData(event, msg, strlen(msg) + 1);
}

static SEXP tk_eval(char *cmd)
{
    ipc_event_t event;
    ipctk_result_code_t code;
    size_t size = strlen(cmd);
    char *data;
    SEXP ans;

    if (IPC_CreateEvent(&event) != 0)
	error("event creation failed");
    if (IPC_PutEventData(event, cmd, size + 1) != 0) {
	IPC_DisposeEvent(event);
	error("adding event data failed");
    }
    if (IPC_SendEvent(sink, event, 1) != 0) {
	IPC_DisposeEvent(event);
	error("send failed");
    }
    if (IPC_GetEventData(event, (void **) &data, &size) != 0) {
	IPC_DisposeEvent(event);
	error("reading event data failed");
    }

    code = *((ipctk_result_code_t *) data);
    data += sizeof(code);
    if (code == IPCTK_ERROR) {
	char p[512];
	if (strlen(data) > 500)
	    strcpy(p, "tcl error.\n");
	else
	    sprintf(p, "[tcl] %s.\n", data);
	IPC_DisposeEvent(event);
	error(p);
    }
    ans = mkString(data);
    IPC_DisposeEvent(event);
    return ans;
}

/* FIXME get rid of fixed size buffers in do_Tclcallback et al. (low
   priority, since there is no meaningful way to override the current
   buffer length) */

static SEXP dotTcl(SEXP args)
{
    if(!isValidString(CADR(args)))
	error("invalid argument");
    return tk_eval(CHAR(STRING_ELT(CADR(args), 0)));
}

/* Setup to invoke callback from Tcl. Notice that something needs to
   ensure that the callback is protected from GC as long as anything
   is around to invoke it. This is handled in interpreted code by
   assigning into the environment of the window with which the
   callback is associated */

static SEXP dotTclcallback(SEXP args)
{
    SEXP ans, callback = CADR(args);

    char buf[256];

    if (isFunction(callback)) {
	SEXP closure = callback;
	SEXP formals = FORMALS(closure);
	char tmp[20];

	sprintf(buf, "R_call 0x%lx", (unsigned long) closure);

	while ( formals != R_NilValue ) {
	    /***** protect against buffer overflow */
	    if (TAG(formals) ==  R_DotsSymbol) break;
	    sprintf(tmp, " %%%s", CHAR(PRINTNAME(TAG(formals))));
	    strcat(buf, tmp);
	    formals = CDR(formals);
	}
    }
    else if (isLanguage(callback)) {
	SEXP call = callback;
        SEXP env = CADDR(args);
	sprintf(buf, "R_call_lang 0x%lx 0x%lx", 
		(unsigned long) call,
		(unsigned long) env);
    }
    else
	error("argument is not of correct type");

    ans = mkString(buf);
    return ans;
}

static void myHandlerFcn(void *userData)
{
    /**** FIXME: We need a select here because R seems to call the
       handler at simes when there is no activity.  Without the select
       ^Z causes a hang in the read.  Not clear eating all available
       input is the right thing to do--might unfairly favor this event
       source */
    int fd;
    IPC_SinkFD(sink, &fd);

    for (;;) {
	fd_set rfds, wfds, efds;
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = 0;

	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	FD_ZERO(&efds);
	FD_SET(fd, &rfds);

	if (select(fd + 1, &rfds, &wfds, &efds, &tv) > 0 &&
	    FD_ISSET(fd, &rfds)) {
	    if (IPC_HandleEvent(sink) != 0) {
		fprintf(stderr, "event handler failed\n");
		break;
	    }
	}
	else
	    break;
    }
}

static InputHandler *myHandlerToken = NULL;
static int Tcl_loaded = 0;

static void addTcl(void)
{
    int fd;
    IPC_SinkFD(sink, &fd);
    if (Tcl_loaded)
	error("Tcl already loaded");
    Tcl_loaded = 1;
    myHandlerToken = addInputHandler(R_InputHandlers, fd, myHandlerFcn, -1);
    if (myHandlerToken == NULL)
	error("Could not install input handler"); /**** need cleanup */
}

static void delTcl(void)
{
    if (!Tcl_loaded)
	error("Tcl is not loaded");
    removeInputHandler(&R_InputHandlers, myHandlerToken);
    myHandlerToken = NULL;
    Tcl_loaded = 0;
}

static void ipctcltk_start(char **path)
{
    char *argv[5];

    argv[0] = "xyzq";
    argv[1] = "R_eval";
    argv[2] = "R_call";
    argv[3] = "R_call_lang";
    argv[4] = NULL;

    if (IPC_StartServer(*path, argv, &sink, handler) != 0)
	error("connect failed\n");
    addTcl();
}

#include <R_ext/Rdynload.h>

static const R_CMethodDef CEntries[] = {
    {"ipctcltk_start", (DL_FUNC) &ipctcltk_start, 1},
    {"delIpcTcl", (DL_FUNC) &delTcl, 0},
    {NULL, NULL, 0}
};

static const R_ExternalMethodDef ExternEntries[] = {
    {"dotIpcTcl", (DL_FUNC) &dotTcl, -1},
    {"dotIpcTclcallback", (DL_FUNC) &dotTclcallback, -1},
    {NULL, NULL, 0}
};


void R_init_ipctcltk(DllInfo *dll)
{
    R_registerRoutines(dll, CEntries, NULL, NULL, ExternEntries);
}
